import numpy as np
import gegd.parameter_processing.symmetry_operations as symOp
import gegd.parameter_processing.density_transforms as dtf
import time
import os

from mpi4py import MPI
comm = MPI.COMM_WORLD

class optimizer:
    def __init__(self,
                 Nx,
                 Ny,
                 symmetry,
                 periodic,
                 padding,
                 maxiter,
                 high_fidelity_setting,
                 brush_size,
                 upsample_ratio=1,
                 beta_proj=8,
                 brush_shape='circle',
                 cost_obj=None,
                 Nthreads=1,
                 ):
        
        self.Nx = Nx
        self.Ny = Ny
        self.symmetry = symmetry
        self.periodic = periodic
        self.padding = padding
        self.maxiter = maxiter
        self.beta_proj = beta_proj
        self.beta_proj_sigma = beta_proj
        self.brush_size = brush_size
        self.upsample_ratio = upsample_ratio
        self.brush_shape = brush_shape
        self.sigma_filter = brush_size/2/np.sqrt(2)
        self.high_fidelity_setting = high_fidelity_setting
        self.cost_obj = cost_obj
        self.Nthreads = Nthreads
        
        # Get Number of Independent Parameters
        if symmetry == 0:
            self.Ndim = Nx*Ny
           
        elif symmetry == 1:
            self.Ndim = int(np.floor(Nx/2 + 0.5)*Ny)
        
        elif symmetry == 2:
            self.Ndim = int(np.floor(Nx/2 + 0.5)*np.floor(Ny/2 + 0.5))
        
        elif symmetry == 4:
            self.Ndim = int(np.floor(Nx/2 + 0.5)*(np.floor(Nx/2 + 0.5) + 1)/2)
    
    def get_cost_samples(self, x):
        # Get Brush Binarized Densities ------------------------------------------------------------
        x_bin = dtf.binarize(x, self.symmetry, self.periodic, self.Nx, self.Ny, self.brush_size, self.brush_shape, self.beta_proj, self.sigma_filter,
                             upsample_ratio=self.upsample_ratio, padding=self.padding, Nthreads=self.Nthreads)

        # Sample Modified Cost Function --------------------------------------------------------------
        self.cost_obj.set_accuracy(self.high_fidelity_setting)
        
        f_batch = np.zeros(self.Nsample)
        for n in range(self.Nsample):
            f_batch[n] = self.cost_obj.get_cost(x_bin[n,:], get_grad=False)
        
        return f_batch, x_bin

    def CMA_ES(
        self,
        xmean=None,
        sigma=None,
        Cov=None,
        p_c=None,
        p_sigma=None,
        B=None,
        D=None,
        eig_eval=None,
        counteval=None,
    ):
        # Set Selection Parameters
        self.Nsample = 25 #int(4 + np.floor(3 * np.log(self.Ndim)))
        mu = int(np.floor(self.Nsample / 2))
        weights = np.log(mu + 1 / 2) - np.log(np.arange(mu) + 1)
        weights /= np.sum(weights)
        mu_eff = 1 / np.sum(weights**2)

        # Set Adaptation Parameters
        c_c = (4 + mu_eff / self.Ndim) / (self.Ndim + 4 + 2 * mu_eff / self.Ndim)
        c_sigma = (mu_eff + 2) / (self.Ndim + mu_eff + 5)
        c_1 = 2 / ((self.Ndim + 1.3)**2 + mu_eff)
        c_mu = np.min((1 - c_1, 2 * (mu_eff - 2 + 1 / mu_eff) / ((self.Ndim + 2)**2 + 2 * mu_eff / 2)))
        d_sigma = 1 + 2 * np.max((0, np.sqrt((mu_eff - 1) / (self.Ndim + 1)) - 1)) + c_sigma

        # Initialize Dynamic Internal Parameters
        p_c = p_c if p_c is not None else np.zeros(self.Ndim)
        p_sigma = p_sigma if p_sigma is not None else np.zeros(self.Ndim)
        B = B if B is not None else np.identity(self.Ndim)
        D = D if D is not None else np.identity(self.Ndim)
        xmean = xmean if xmean is not None else np.zeros(self.Ndim)
        sigma = sigma if sigma is not None else 0.5
        Cov = Cov if Cov is not None else np.identity(self.Ndim)
        eig_eval = eig_eval if eig_eval is not None else 0
        chi_N = self.Ndim**0.5 * (1 - 1 / (4 * self.Ndim) + 1 / (21 * self.Ndim**2))

        counteval = counteval if counteval is not None else 0

        while True:
            t1 = time.time()

            # Sample Candidates
            arz = np.random.normal(size=(self.Ndim, self.Nsample))
            arx = xmean[:,np.newaxis] + sigma * (B @ D @ arz)
            
            # Cost Evaluation
            cost, x_bin = self.get_cost_samples(arx.T)
            sorted_index = np.argsort(cost)
            cost = cost[sorted_index]
            arx = arx[:,sorted_index]
            x_bin = x_bin[sorted_index,:]
            counteval += self.Nsample

            if self.x_mean_hist is None:
                self.x_mean_hist = xmean.copy()
            else:
                self.x_mean_hist = np.vstack((self.x_mean_hist, xmean))

            if self.x_latent_hist is None:
                self.x_latent_hist = arx[:,0].copy()
            else:
                self.x_latent_hist = np.vstack((self.x_latent_hist, arx[:,0]))
            
            if self.best_cost_hist.size == 0:
                new_best = True
                self.best_cost_hist = np.append(self.best_cost_hist, cost[0])
            else:
                new_best = self.best_cost_hist[-1] > cost[0]
                self.best_cost_hist = np.append(self.best_cost_hist, np.min((self.best_cost_hist[-1], cost[0])))
                
            if self.best_x_hist is None:
                self.best_x_hist = x_bin[0,:].copy()
            else:
                if new_best:
                    self.best_x_hist = np.vstack((self.best_x_hist, x_bin[0,:]))
                else:
                    if self.best_x_hist.ndim == 1:
                        self.best_x_hist = np.vstack((self.best_x_hist, self.best_x_hist))
                    else:
                        self.best_x_hist = np.vstack((self.best_x_hist, self.best_x_hist[-1,:]))
            
            self.cost_ensemble_hist = np.append(self.cost_ensemble_hist, np.mean(cost))
            self.cost_ensemble_sigma_hist = np.append(self.cost_ensemble_sigma_hist, np.std(cost))
            self.sigma_hist = np.append(self.sigma_hist, sigma)
            
            t2 = time.time()
            t_rem = (t2 - t1)*(self.maxiter - self.n_iter + 1)/3600
            
            if comm.rank == 0:
                print('    | %4d |  %4d | %7.5f |  %9.2f |  %8.3f |   %5.2f   |' %(
                    self.n_iter,
                    self.Nsample,
                    sigma,
                    np.mean(cost),
                    self.best_cost_hist[-1],
                    t_rem,
                ), flush=True)

            self.save_data(
                p_c=p_c,
                p_sigma=p_sigma,
                B=B,
                D=D,
                xmean=xmean,
                sigma=sigma,
                Cov=Cov,
                eig_eval=eig_eval,
                counteval=counteval,
            )

            if self.n_iter > self.maxiter:
                self.n_iter += 1
                break

            # Update Mean
            xmean = arx[:,:mu] @ weights
            zmean = arz[:,:mu] @ weights

            # Update Evolution Path and Step Size
            p_sigma = (1 - c_sigma) * p_sigma + np.sqrt(c_sigma * (2 - c_sigma) * mu_eff) * (B @ zmean)
            h_sigma = np.linalg.norm(p_sigma) / np.sqrt(1 - (1 - c_sigma)**(2 * counteval / self.Nsample)) / chi_N < 1.4 + 2 / (self.Ndim + 1)
            p_c = (1 - c_c) * p_c + h_sigma * np.sqrt(c_c * (2 - c_c) * mu_eff) * (B @ D @ zmean)

            # Update Covariance Matrix
            Cov = (1 - c_1 - c_mu) * Cov \
                + c_1 * (p_c @ p_c.T + (1 - h_sigma) * c_c * (2 - c_c) * Cov) \
                + c_mu * (B @ D @ arz[:,:mu]) @ np.diag(weights) @ (B @ D @ arz[:,:mu]).T
            
            # Update Step Size
            sigma = sigma * np.exp((c_sigma / d_sigma) * (np.linalg.norm(p_sigma) / chi_N - 1))

            # Update B and D
            if counteval - eig_eval > self.Nsample / ((c_1 + c_mu) * self.Ndim * 10):
                eig_eval = counteval
                Cov = np.triu(Cov) + np.triu(Cov, k=1).T
                [D, B] = np.linalg.eigh(Cov)
                D = np.sqrt(np.diag(D))
            
            self.n_iter += 1

    def run(self, n_seed, output_filename, x0=None, load_data=False):
        if comm.rank == 0:
            print('### CMA-ES (seed = ' + str(n_seed) + ')\n', flush=True)
    
        np.random.seed(n_seed)
        self.output_filename = output_filename
        
        if load_data:
            data_file1 = output_filename + "_CMA_ES_results.npz"
            data_file2 = output_filename + "_CMA_ES_density_hist.npz"
                
            with np.load(data_file1) as data:
                self.n_iter = data['n_iter']
                
                self.cost_ensemble_hist = data['cost_ensemble_hist'][:self.n_iter]
                self.best_cost_hist = data['best_cost_hist'][:self.n_iter]
                self.cost_ensemble_sigma_hist = data['cost_ensemble_sigma_hist'][:self.n_iter]
                self.sigma_hist = data['sigma_hist'][:self.n_iter]

                sigma = data['sigma']
                eig_eval = data['eig_eval']
                counteval = data['counteval']
                
            with np.load(data_file2) as data:
                self.x_latent_hist = data['x_latent_hist'][:self.n_iter,:]
                self.x_mean_hist = data['x_mean_hist'][:self.n_iter,:]
                self.best_x_hist = data['best_x_hist'][:self.n_iter,:]
                
                p_c = data['p_c']
                p_sigma = data['p_sigma']
                B = data['B']
                D = data['D']
                x0 = data['xmean']
                Cov = data['Cov']

        else:
            self.n_iter = 0
            
            self.x_latent_hist = None
            self.x_mean_hist = None
            self.best_x_hist = None
            self.cost_ensemble_hist = np.zeros(0)
            self.best_cost_hist = np.zeros(0)
            self.cost_ensemble_sigma_hist = np.zeros(0)
            self.sigma_hist = np.zeros(0)
            
            if x0 is None:
                # Initial Structure
                x0 = np.zeros(self.Ndim)
            sigma = None
            Cov = None
            p_c = None
            p_sigma = None
            B = None
            D = None
            eig_eval = None
            counteval = None
        
        if comm.rank == 0:
            print('    | Iter |   N   |  sigma  |  Cost Ens  | Cost Best | t_rem(hr) |',
                  flush=True)

        self.CMA_ES(
            xmean=x0,
            sigma=sigma,
            Cov=Cov,
            p_c=p_c,
            p_sigma=p_sigma,
            B=B,
            D=D,
            eig_eval=eig_eval,
            counteval=counteval,
        )
        
        self.save_data()
    
    def save_data(
        self,
        p_c=None,
        p_sigma=None,
        B=None,
        D=None,
        xmean=None,
        sigma=None,
        Cov=None,
        eig_eval=None,
        counteval=None,
    ):

        if comm.rank == 0:
            if p_c is None:
                with np.load(self.output_filename + "_CMA_ES_density_hist.npz") as data:
                    p_c = data['p_c']
                    p_sigma = data['p_sigma']
                    B = data['B']
                    D = data['D']
                    xmean = data['xmean']
                    Cov = data['Cov']

            if sigma is None:
                with np.load(self.output_filename + "_CMA_ES_results.npz") as data:
                    sigma = data['sigma']
                    eig_eval = data['eig_eval']
                    counteval = data['counteval']
        
            # Customize below
            np.savez(self.output_filename + "_CMA_ES_results",
                cost_ensemble_hist=self.cost_ensemble_hist,
                cost_ensemble_sigma_hist=self.cost_ensemble_sigma_hist,
                best_cost_hist=self.best_cost_hist,
                sigma_hist=self.sigma_hist,
                n_iter=self.n_iter,
                sigma=sigma,
                eig_eval=eig_eval,
                counteval=counteval,
            )
                        
            np.savez(self.output_filename + "_CMA_ES_density_hist",
                x_mean_hist=self.x_mean_hist,
                x_latent_hist=self.x_latent_hist,
                best_x_hist=self.best_x_hist,
                p_c=p_c,
                p_sigma=p_sigma,
                B=B,
                D=D,
                xmean=xmean,
                Cov=Cov,
            )


import os
directory = os.path.dirname(os.path.realpath(__file__))

import numpy as np
from scipy.ndimage import convolve, zoom, gaussian_filter, maximum_filter
from numba import jit
import time

def make_feasible(
    weight,
    N_x,
    N_y,
    brush_size,
    brush_type,
    padding=None,
    periodic=True,
    symmetry='None',
    dim=2,
    upsample_ratio=1,
):
    # Reshape weight into a 1D or 2D array
    if dim == 1:
        weight_sq = weight.reshape(-1,1)
    else:
        weight_sq = weight.reshape(N_x, N_y)
    
    N_x = int(N_x*upsample_ratio)
    N_y = int(N_y*upsample_ratio)
    
    if upsample_ratio > 1:
        brush_size = int(np.floor(upsample_ratio*np.sqrt((brush_size - 2)**2 + 1) + 2))
        
        if periodic:
            weight_sq = zoom(weight_sq, upsample_ratio, order=0, mode='wrap')
        else:
            weight_sq = zoom(weight_sq, upsample_ratio, order=0, mode='grid-constant')
        weight_sq = apply_symmetry(symmetry, weight_sq, N_x, N_y)
    
    if not periodic:
        pad_value = -1 if padding == 'void' else (1 if padding == 'solid' else 0)

        # Padding weight_sq and creating the mask for padded regions
        if dim == 1:
            weight_sq = np.pad(weight_sq, (brush_size, brush_size), constant_values=pad_value)

            # Create a mask to identify padded regions for 1D
            mask = np.ones(weight_sq.shape, dtype=bool)
            mask[brush_size:-brush_size] = False
        else:
            weight_sq = np.pad(weight_sq, ((brush_size, brush_size), (brush_size, brush_size)), constant_values=pad_value)

            # Create a mask to identify padded regions for 2D
            mask = np.ones(weight_sq.shape, dtype=bool)
            mask[brush_size:-brush_size, brush_size:-brush_size] = False
    else:
        mask = np.zeros(weight_sq.shape, dtype=bool)  # No padded regions if periodic
    
    # Initialize the void and solid arrays
    pix_solid = np.zeros(weight_sq.shape)
    pix_void = np.zeros(weight_sq.shape)

    # Initialize the touch arrays
    touch_void = np.zeros(weight_sq.shape)
    touch_solid = np.zeros(weight_sq.shape)
    touch_void_pos = np.zeros(weight_sq.shape)
    touch_void_neg = np.zeros(weight_sq.shape)
    touch_solid_pos = np.zeros(weight_sq.shape)
    touch_solid_neg = np.zeros(weight_sq.shape)
    
    # Define a mask for the brush
    r_c = int((brush_size - 1)/2)
    r_c_plus2 = int((brush_size + 1)/2)
    r_conv = int(brush_size + 2*r_c)
    r_conv2 = int(brush_size + 4*r_c)
    if dim == 1:
        brush_shape = np.ones(brush_size).astype(int)
        brush_shape_plus2 = np.ones(brush_size + 6).astype(int)
        brush_conv_shape = np.ones(r_conv)
        brush_conv_shape2 = np.ones(r_conv2)
    elif dim == 2:
        x, y = np.meshgrid(np.arange(brush_size), np.arange(brush_size))
        x_plus2, y_plus2 = np.meshgrid(np.arange(brush_size + 2), np.arange(brush_size + 2))
        if brush_type == 'circle':
            distance = np.sqrt((x - r_c) ** 2 + (y - r_c) ** 2)
            brush_shape = (distance <= r_c).astype(int)
            
            distance = np.sqrt((x_plus2 - r_c_plus2) ** 2 + (y_plus2 - r_c_plus2) ** 2)
            brush_shape_plus2 = (distance <= r_c_plus2).astype(int)
        elif brush_type == 'notched_sq':
            brush_shape = np.ones((brush_size, brush_size)).astype(int)
            brush_shape[0,0] = 0
            brush_shape[0,-1] = 0
            brush_shape[-1,0] = 0
            brush_shape[-1,-1] = 0
        
        brush_conv_shape = np.zeros((r_conv, r_conv))
        brush_conv_shape[r_c:-r_c,r_c:-r_c] = brush_shape
        brush_conv_shape = convolve(brush_conv_shape, brush_shape, mode='constant', cval=0)
        brush_conv_shape = np.where(brush_conv_shape > 0, 1, 0)
        
        brush_conv_shape2 = np.zeros((r_conv2, r_conv2))
        brush_conv_shape2[r_c:-r_c,r_c:-r_c] = brush_conv_shape
        brush_conv_shape2 = convolve(brush_conv_shape2, brush_shape, mode='constant', cval=0)
        brush_conv_shape2 = np.where(brush_conv_shape2 > 0, 1, 0)

    delta_array = np.zeros_like(weight_sq)
    delta_array[0,0] = 1
    reference_convolution0 = convolve(delta_array, brush_shape, mode='wrap').astype(bool)
    reference_convolution1 = convolve(delta_array, brush_conv_shape, mode='wrap').astype(bool)
    reference_convolution2 = convolve(delta_array, brush_conv_shape2, mode='wrap').astype(bool)

    # Compute the reward array
    if periodic:
        score_solid = convolve(weight_sq, brush_shape, mode='wrap')
    else:
        score_solid = convolve(weight_sq, brush_shape, mode='constant', cval=pad_value)
    score_void = -score_solid.copy()
    
    # Pre-fill touches and pixels
    score_solid_pre_fill = convolve(weight_sq, brush_shape_plus2, mode='wrap')
    touch_solid_pos[score_solid_pre_fill>=np.sum(brush_shape_plus2)] = 1
    touch_void_neg[score_solid_pre_fill>=np.sum(brush_shape_plus2)] = 1
    
    touch_solid_neg[score_solid_pre_fill<=-np.sum(brush_shape_plus2)] = 1
    touch_void_pos[score_solid_pre_fill<=-np.sum(brush_shape_plus2)] = 1
    
    if periodic:
        pix_solid += convolve(touch_solid_pos.astype(int), brush_shape, mode='wrap').astype(bool).astype(int)
        touch_solid += touch_solid_pos
        touch_void -= convolve(touch_solid_pos.astype(int), brush_conv_shape, mode='wrap').astype(bool).astype(int)
        
        pix_solid -= convolve(touch_solid_neg.astype(int), brush_shape, mode='wrap').astype(bool).astype(int)
        touch_void += touch_solid_neg
        touch_solid -= convolve(touch_solid_neg.astype(int), brush_conv_shape, mode='wrap').astype(bool).astype(int)
    else:
        pix_solid += convolve(touch_solid_pos.astype(int), brush_shape, mode='constant').astype(bool).astype(int)
        touch_solid += touch_solid_pos
        touch_void -= convolve(touch_solid_pos.astype(int), brush_conv_shape, mode='constant').astype(bool).astype(int)
        
        pix_solid -= convolve(touch_solid_neg.astype(int), brush_shape, mode='constant').astype(bool).astype(int)
        touch_void += touch_solid_neg
        touch_solid -= convolve(touch_solid_neg.astype(int), brush_conv_shape, mode='wrap').astype(bool).astype(int)
    
    if np.sum(pix_solid==0) > 0:
        try:
            pix_solid = main_loop(score_solid,
                                  score_void,
                                  pix_solid,
                                  pix_void,
                                  touch_solid,
                                  touch_void,
                                  reference_convolution0,
                                  reference_convolution1,
                                  reference_convolution2,
                                  symmetry,
                                  periodic,
                                  brush_size,
                                  dim,
                                  N_x,
                                  N_y,
                                  )
        except:
            np.savez(directory + '/debug_make_feasible', weight=weight)
            assert False

    pix_solid[pix_solid<0] = 0
    if not periodic:
        if dim == 1:
            pix_solid = pix_solid[brush_size:-brush_size]
        else:
            pix_solid = pix_solid[brush_size:-brush_size,brush_size:-brush_size]
    
    return pix_solid.reshape(-1)

def find_index_max(symmetry, periodic, brush_size, dim, N_x, N_y, score, touch, max_value):
    ind_temp = np.argwhere(np.abs(score - max_value) < 1e-4)
    if symmetry == 'None':
        for i in range(ind_temp.shape[0]):
            if dim == 1:
                if touch[ind_temp[i,0]] == 0:
                    return ind_temp[i,0]
            else:
                if touch[ind_temp[i,0],ind_temp[i,1]] == 0:
                    return tuple(ind_temp[i,:])
    
    elif symmetry == 'D1':
        half_x = np.floor(N_x/2 + 0.5) + 2*(1 - periodic)*brush_size
        
        for i in range(ind_temp.shape[0]):
            if ind_temp[i,0] <= half_x:
                if touch[ind_temp[i,0],ind_temp[i,1]] == 0:
                    return tuple(ind_temp[i,:])
    
    elif symmetry == 'D2':
        half_x = np.floor(N_x/2 + 0.5) + 2*(1 - periodic)*brush_size
        half_y = np.floor(N_y/2 + 0.5) + 2*(1 - periodic)*brush_size
        
        for i in range(ind_temp.shape[0]):
            if ind_temp[i,0] <= half_x and ind_temp[i,1] <= half_y:
                if touch[ind_temp[i,0],ind_temp[i,1]] == 0:
                    return tuple(ind_temp[i,:])
    
    elif symmetry == 'D4':
        half_y = np.floor(N_y/2 + 0.5) + 2*(1 - periodic)*brush_size
        
        for i in range(ind_temp.shape[0]):
            if ind_temp[i,0] <= ind_temp[i,1] and ind_temp[i,1] <= half_y:
                if touch[ind_temp[i,0],ind_temp[i,1]] == 0:
                    return tuple(ind_temp[i,:])

def main_loop(score_solid,
              score_void,
              pix_solid,
              pix_void,
              touch_solid,
              touch_void,
              reference_convolution0,
              reference_convolution1,
              reference_convolution2,
              symmetry,
              periodic,
              brush_size,
              dim,
              N_x,
              N_y,
              ):
    
    n_iter = 0
    while True:
        t1 = time.time()
        solid = False
        void = False
        
        max_solid, max_void = -1e10, -1e10

        # Find max scores for solid and void where touches have not been applied
        if np.sum(touch_solid == 0) > 0:
            max_solid = np.nanmax(score_solid[touch_solid == 0])
        
        if np.sum(touch_void == 0) > 0:
            max_void = np.nanmax(score_void[touch_void == 0])
        
        if max_solid > max_void:
            solid = True
            ind_max = find_index_max(symmetry, periodic, brush_size, dim, N_x, N_y, score_solid, touch_solid, max_solid)
        elif max_void > max_solid:
            void = True
            ind_max = find_index_max(symmetry, periodic, brush_size, dim, N_x, N_y, score_void, touch_void, max_void)
        else:
            # Random choice if max_solid == max_void
            if np.random.rand() > 0.5:
                solid = True
                ind_max = find_index_max(symmetry, periodic, brush_size, dim, N_x, N_y, score_solid, touch_solid, max_solid)
            else:
                void = True
                ind_max = find_index_max(symmetry, periodic, brush_size, dim, N_x, N_y, score_void, touch_void, max_void)
        if ind_max is None:
            pass
        
        n_iter += 1
        n_empty = np.sum(pix_solid==0)
        touch_solid, touch_void, pix_solid, pix_void = make_touch(solid,
                                                                  void,
                                                                  ind_max,
                                                                  touch_solid,
                                                                  touch_void,
                                                                  pix_solid,
                                                                  pix_void,
                                                                  reference_convolution0,
                                                                  reference_convolution1,
                                                                  symmetry,
                                                                  )
        
        if n_empty != np.sum(pix_solid==0):
            touch_solid, touch_void, pix_solid, pix_void = fill_required_pixels(ind_max,
                                                                                touch_solid,
                                                                                touch_void,
                                                                                pix_solid,
                                                                                pix_void,
                                                                                score_solid,
                                                                                score_void,
                                                                                reference_convolution0,
                                                                                reference_convolution1,
                                                                                reference_convolution2,
                                                                                symmetry,
                                                                                )
        
        t2 = time.time()
        if n_iter%10 == 0:
            print('Iter ' + str(n_iter) + ' | Unassigned Solid Pixels: ' + str(np.sum(pix_solid==0)) + ' | Unassigned Solid Touches: ' + str(np.sum(touch_solid==0)) + ' | Unassigned Void Touches: ' + str(np.sum(touch_void==0)) + ' | Time: ' + str(t2 - t1), flush=True)
        if np.sum(pix_solid==0) == 0:
            break
    
    return pix_solid

def fill_required_pixels(ind_max,
                         touch_solid,
                         touch_void,
                         pix_solid,
                         pix_void,
                         score_solid,
                         score_void,
                         reference_convolution0,
                         reference_convolution1,
                         reference_convolution2,
                         symmetry,
                         ):
    
    last_affected = np.roll(reference_convolution2, ind_max, axis=(0,1)).astype(bool)
    
    required = True
    while required:
        required = False
        
        pix_to_check = np.argwhere((pix_solid == 0)*last_affected)
        for n in pix_to_check:
            # if n[0] > 109 and n[0] < 114 and n[1] > 14 and n[1] < 24:
            #     pass
            required_mask = np.roll(reference_convolution0, n, axis=(0,1))
            
            if np.all(touch_solid[required_mask]):
                score_required = np.where(required_mask*(touch_void==0), score_void, np.nan)
                if not np.isnan(np.nanmax(score_required)):
                    ind_required_max = np.argwhere(score_required == np.nanmax(score_required))[0,:].astype(int)

                    touch_solid, touch_void, pix_solid, pix_void = make_touch(False,
                                                                              True,
                                                                              ind_required_max,
                                                                              touch_solid,
                                                                              touch_void,
                                                                              pix_solid,
                                                                              pix_void,
                                                                              reference_convolution0,
                                                                              reference_convolution1,
                                                                              symmetry,
                                                                              )
                    
                    last_affected += np.roll(reference_convolution2, ind_required_max, axis=(0,1))
                    
                    required = True
                    break
                else:
                    pass
                
            elif np.all(touch_void[required_mask]):
                score_required = np.where(required_mask*(touch_solid==0), score_solid, np.nan)
                if not np.isnan(np.nanmax(score_required)):
                    ind_required_max = np.argwhere(score_required == np.nanmax(score_required))[0,:].astype(int)

                    touch_solid, touch_void, pix_solid, pix_void = make_touch(True,
                                                                              False,
                                                                              ind_required_max,
                                                                              touch_solid,
                                                                              touch_void,
                                                                              pix_solid,
                                                                              pix_void,
                                                                              reference_convolution0,
                                                                              reference_convolution1,
                                                                              symmetry,
                                                                              )
                    
                    last_affected += np.roll(reference_convolution2, ind_required_max, axis=(0,1))
                    
                    required = True
                    break
                else:
                    pass
            else:
                last_affected[tuple(n)] = 0
        
    return touch_solid, touch_void, pix_solid, pix_void

def make_touch(solid,
               void,
               ind_touch,
               touch_solid,
               touch_void,
               pix_solid,
               pix_void,
               reference_convolution0,
               reference_convolution1,
               symmetry,
               ):
    
    if solid:
        touch_solid, touch_void = touch2pix(ind_touch,
                                            touch_solid,
                                            touch_void,
                                            pix_solid,
                                            pix_void,
                                            reference_convolution1,
                                            symmetry,
                                            )
    elif void:
        touch_void, touch_solid = touch2pix(ind_touch,
                                            touch_void,
                                            touch_solid,
                                            pix_void,
                                            pix_solid,
                                            reference_convolution1,
                                            symmetry,
                                            )

    pix_mask = np.roll(reference_convolution0, ind_touch, axis=(0,1))
    pix_solid[(pix_mask & (pix_solid == 0))] = (-1)**(solid + 1)
    
    if symmetry != 'None':
        pix_solid = apply_symmetry(symmetry, pix_solid, pix_solid.shape[0], pix_solid.shape[1])
    
    return touch_solid, touch_void, pix_solid, pix_void

def touch2pix(ind_touch,
              touch1,
              touch2,
              pix1,
              pix2,
              reference_convolution1,
              symmetry,
              ):
    
    touch_mask = np.roll(reference_convolution1, ind_touch, axis=(0,1))

    # Update touch1 to indicate a successful touch
    touch1[tuple(ind_touch)] = 1
    # Update touch2 to mark areas covered by the touch as invalid (-1)
    touch2[(touch_mask & (touch2 == 0))] = -1

    if symmetry != 'None':
        touch1 = apply_symmetry(symmetry, touch1, touch1.shape[0], touch1.shape[1])
        touch2 = apply_symmetry(symmetry, touch2, touch2.shape[0], touch1.shape[1])

    return touch1, touch2

def apply_symmetry(symmetry, x, N_x, N_y):
    if symmetry == 'None':
        x_sym = x.copy()
        
    elif symmetry == 'D1':
        x_half = x[:int(np.floor(N_x/2 + 0.5)),:]
        x_sym = np.zeros((N_x, N_y))
        
        if N_x % 2 == 0:
            x_sym[:int(N_x/2),:] = x_half
            x_sym[int(N_x/2):,:] = np.flipud(x_half)
        
        else:
            x_sym[:int((N_x+1)/2),:] = x_half
            x_sym[int((N_x+1)/2):,:] = np.flipud(x_half[:-1,:])
    
    elif symmetry == 'D2':
        x_quart = x[:int(np.floor(N_x/2 + 0.5)),:int(np.floor(N_y/2 + 0.5))]
        x_sym = np.zeros((N_x, N_y))
        
        if N_x % 2 == 0 and N_y % 2 == 0:
            x_sym[:int(N_x/2),:int(N_y/2)] = x_quart
            x_sym[int(N_x/2):,:int(N_y/2)] = np.flipud(x_quart)
            x_sym[:int(N_x/2),int(N_y/2):] = np.fliplr(x_quart)
            x_sym[int(N_x/2):,int(N_y/2):] = np.fliplr(np.flipud(x_quart))
        
        elif N_x % 2 == 1 and N_y % 2 == 0:
            x_sym[:int((N_x+1)/2),:int(N_y/2)] = x_quart
            x_sym[int((N_x+1)/2):,:int(N_y/2)] = np.flipud(x_quart[:-1,:])
            x_sym[:int((N_x+1)/2),int(N_y/2):] = np.fliplr(x_quart)
            x_sym[int((N_x+1)/2):,int(N_y/2):] = np.fliplr(np.flipud(x_quart[:-1,:]))
        
        elif N_x % 2 == 0 and N_y % 2 == 1:
            x_sym[:int(N_x/2),:int((N_y+1)/2)] = x_quart
            x_sym[int(N_x/2):,:int((N_y+1)/2)] = np.flipud(x_quart)
            x_sym[:int(N_x/2),int((N_y+1)/2):] = np.fliplr(x_quart[:,:-1])
            x_sym[int(N_x/2):,int((N_y+1)/2):] = np.fliplr(np.flipud(x_quart[:,:-1]))
        
        elif N_x % 2 == 1 and N_y % 2 == 1:
            x_sym[:int((N_x+1)/2),:int((N_y+1)/2)] = x_quart
            x_sym[int((N_x+1)/2):,:int((N_y+1)/2)] = np.flipud(x_quart[:-1,:])
            x_sym[:int((N_x+1)/2),int((N_y+1)/2):] = np.fliplr(x_quart[:,:-1])
            x_sym[int((N_x+1)/2):,int((N_y+1)/2):] = np.fliplr(np.flipud(x_quart[:-1,:-1]))

    elif symmetry == 'D4':
        if N_x % 2 == 0:
            x_triu = x[:int(N_x/2),:int(N_x/2)][np.triu_indices(int(N_x/2))]
            
            x_quart = np.zeros((int(N_x/2), int(N_x/2)))
            x_quart[np.triu_indices(int(N_x/2))] = x_triu
            x_quart = x_quart + x_quart.T - np.diag(np.diag(x_quart))
            
            x_sym = np.zeros((N_x, N_x))
            x_sym[:int(N_x/2),:int(N_x/2)] = x_quart
            x_sym[int(N_x/2):,:int(N_x/2)] = np.rot90(x_quart)
            x_sym[int(N_x/2):,int(N_x/2):] = np.rot90(x_quart, k=2)
            x_sym[:int(N_x/2),int(N_x/2):] = np.rot90(x_quart, k=-1)
        
        else:
            x_triu = x[:int((N_x+1)/2),:int((N_x+1)/2)][np.triu_indices(int((N_x+1)/2))]
            
            x_quart = np.zeros((int((N_x+1)/2), int((N_x+1)/2)))
            x_quart[np.triu_indices(int((N_x+1)/2))] = x_triu
            x_quart = x_quart + x_quart.T - np.diag(np.diag(x_quart))
            
            x_sym = np.zeros((N_x, N_x))
            x_sym[:int((N_x+1)/2),:int((N_x+1)/2)] = x_quart
            x_sym[int((N_x+1)/2):,:int((N_x+1)/2)] = np.rot90(x_quart)[1:,:]
            x_sym[int((N_x+1)/2):,int((N_x+1)/2):] = np.rot90(x_quart, k=2)[1:,1:]
            x_sym[:int((N_x+1)/2),int((N_x+1)/2):] = np.rot90(x_quart, k=-1)[:,1:]
    
    return x_sym


import numpy as np
import os
from scipy.ndimage import gaussian_filter
import torch
import gegd.parameter_processing.symmetry_operations as symOp
import gegd.parameter_processing.feasible_design_generator.fdg as FDG
import gegd.parameter_processing.bilevel_design_generator as bdg
import time

from mpi4py import MPI
comm = MPI.COMM_WORLD

def filter_and_project(x, symmetry, periodic, Nx, Ny, brush_size, sigma_filter, beta_proj, padding=None):
    x_sym = symOp.symmetrize(x, symmetry, Nx, Ny)

    if sigma_filter is not None:
        if periodic:
            x_filter = gaussian_filter(x_sym, sigma=sigma_filter, mode='wrap')
        else:
            padding[brush_size:-brush_size,brush_size:-brush_size] = x_sym.copy()
            x_filter = gaussian_filter(padding, sigma=sigma_filter, mode='constant', cval=0)[brush_size:-brush_size,brush_size:-brush_size]
    else:
        x_filter = x_sym.copy()
    
    x_desym = symOp.desymmetrize(x_filter, symmetry, Nx, Ny)
    
    if beta_proj == np.inf:
        x_proj = x_desym.copy()
        x_proj[x_desym<=0] = -1
        x_proj[x_desym>0] = 1
    else:
        x_proj = np.tanh(beta_proj*x_desym)
    
    return x_proj
    
def backprop_filter_and_project(jac_sym, x_latent, symmetry, periodic, Nx, Ny, brush_size, sigma_filter, beta_proj, padding=None):
    x_sym = symOp.symmetrize(x_latent, symmetry, Nx, Ny)

    if sigma_filter is not None:
        if periodic:
            x_filter = gaussian_filter(x_sym, sigma=sigma_filter, mode='wrap')
        else:
            padding[brush_size:-brush_size,brush_size:-brush_size] = x_sym.copy()
            x_filter = gaussian_filter(padding, sigma=sigma_filter, mode='constant', cval=0)[brush_size:-brush_size,brush_size:-brush_size]
    else:
        x_filter = x_sym.copy()

    jac_proj = beta_proj/np.cosh(beta_proj*x_filter)**2
    jac_proj *= jac_sym
    
    if sigma_filter is not None:
        if periodic:
            jac_filter = gaussian_filter(jac_proj, sigma=sigma_filter, mode='wrap')
        else:
            jac_filter = gaussian_filter(jac_proj, sigma=sigma_filter, mode='constant', cval=0)
        
    jac_desym = symOp.desymmetrize_jacobian(jac_filter, symmetry, Nx, Ny)
    
    return jac_desym

def binarize(x, symmetry, periodic, Nx, Ny, brush_size, brush_shape, beta_proj, sigma_filter, dx=None, upsample_ratio=1, padding=None, output_details=False, Nthreads=1, print_runtime_details=False,):
    if dx is not None:
        N_designs = dx.shape[0]
    else:
        N_designs = x.shape[0]

    if padding is None:
        x_sym = np.zeros((N_designs, Nx, Ny)).astype(np.float32)
    else:
        x_sym = np.zeros((N_designs, Nx + 2*brush_size, Ny + 2*brush_size)).astype(np.float32)

    for n in range(N_designs):
        if dx is not None:
            x_fp = filter_and_project(x, symmetry, periodic, Nx, Ny, brush_size, sigma_filter, beta_proj, padding=padding).reshape(-1) + dx[n,:]
        else:
            x_fp = filter_and_project(x[n,:], symmetry, periodic, Nx, Ny, brush_size, sigma_filter, beta_proj, padding=padding)
    
        if padding is None:
            x_sym[n,:,:] = symOp.symmetrize(x_fp, symmetry, Nx, Ny)
        else:
            x_sym[n,:,:] = padding.copy()
            x_sym[n,brush_size:-brush_size,brush_size:-brush_size] = symOp.symmetrize(x_fp, symmetry, Nx, Ny)
    
    t1 = time.time()    
    if brush_size is not None:
        if Nthreads > 0:
            if upsample_ratio == 1:
                x_brush = FDG.make_feasible_parallel(
                    x_sym,
                    brush_size,
                    periodic,
                    symmetry,
                    2,
                    upsample_ratio,
                    Nthreads).reshape(N_designs, -1)
                
            else:
                x_brush_lowres = FDG.make_feasible_parallel(
                    x_sym,
                    brush_size,
                    periodic,
                    symmetry,
                    2,
                    1,
                    Nthreads).reshape(N_designs, Nx, Ny)
                
                x_brush = FDG.make_feasible_parallel(
                    2*x_brush_lowres.astype(np.float32)-1,
                    brush_size,
                    periodic,
                    symmetry,
                    2,
                    upsample_ratio,
                    Nthreads).reshape(N_designs, -1)
        
        else:
            quo, rem = divmod(N_designs, comm.size)
            data_size = np.array([quo + 1 if p < rem else quo for p in range(comm.size)]).astype(np.int32)
            data_disp = np.array([sum(data_size[:p]) for p in range(comm.size + 1)])
            x_sym_proc = x_sym[data_disp[comm.rank]:data_disp[comm.rank+1],:,:]
            
            if data_size[comm.rank] > 0:
                if upsample_ratio == 1:
                    x_brush_proc = FDG.make_feasible_parallel(
                        x_sym_proc,
                        brush_size,
                        periodic,
                        symmetry,
                        2,
                        upsample_ratio,
                        1).reshape(data_size[comm.rank], -1).astype(np.float64)
                    
                else:
                    x_brush_lowres_proc = FDG.make_feasible_parallel(
                        x_sym_proc,
                        brush_size,
                        periodic,
                        symmetry,
                        2,
                        1,
                        1).reshape(data_size[comm.rank], Nx, Ny)
                    
                    x_brush_proc = FDG.make_feasible_parallel(
                        2*x_brush_lowres_proc.astype(np.float32)-1,
                        brush_size,
                        periodic,
                        symmetry,
                        2,
                        upsample_ratio,
                        1).reshape(data_size[comm.rank], -1).astype(np.float64)
            else:
                x_brush_proc = np.array([]).astype(np.float64)
            
            if padding is None:
                data_size_temp = data_size*Nx*Ny*upsample_ratio**2
                data_disp_temp = np.array([sum(data_size_temp[:p]) for p in range(comm.size)])
                
                x_brush_temp = np.zeros(N_designs*Nx*Ny*upsample_ratio**2)
                comm.Allgatherv(x_brush_proc.reshape(-1), [x_brush_temp, data_size_temp, data_disp_temp, MPI.DOUBLE])
                x_brush = x_brush_temp.reshape(N_designs, -1)
            else:
                data_size_temp = data_size*(Nx + 2*brush_size)*(Ny + 2*brush_size)*upsample_ratio**2
                data_disp_temp = np.array([sum(data_size_temp[:p]) for p in range(comm.size)])
                
                x_brush_temp = np.zeros(N_designs*(Nx + 2*brush_size)*(Ny + 2*brush_size)*upsample_ratio**2)
                comm.Allgatherv(x_brush_proc.reshape(-1), [x_brush_temp, data_size_temp, data_disp_temp, MPI.DOUBLE])
                x_brush = x_brush_temp.reshape(N_designs, -1)
    
    t2 = time.time()
    
    if padding is not None:
        x_brush_crop = x_brush.reshape(N_designs, (Nx + 2*brush_size)*upsample_ratio, (Ny + 2*brush_size)*upsample_ratio)
        x_brush_crop = x_brush_crop[:,brush_size:-brush_size,brush_size:-brush_size]
        x_brush = x_brush_crop.reshape(N_designs, -1)
    
    if output_details:
        return np.squeeze(x_sym), np.squeeze(x_brush)
    
    else:
        return np.squeeze(x_brush)

def binarize_arx(
    x,
    symmetry,
    periodic,
    Nx,
    Ny,
    brush_size,
    brush_shape,
    beta_proj,
    sigma_filter,
    dx=None,
    upsample_ratio=1,
    padding=None,
    output_details=False,
    Nthreads=1,
    print_runtime_details=False,
):
    if dx is not None:
        N_designs = dx.shape[0]
        x_reward = np.zeros_like(dx).astype(np.float64)
    else:
        N_designs = x.shape[0]
        x_reward = np.zeros_like(x).astype(np.float64)
    
    for n in range(N_designs):
        if dx is not None:
            x_reward[n,:] = filter_and_project(x, symmetry, periodic, Nx, Ny, brush_size, sigma_filter, beta_proj, padding=padding).reshape(-1) + dx[n,:]
        else:
            x_reward[n,:] = filter_and_project(x[n,:], symmetry, periodic, Nx, Ny, brush_size, sigma_filter, beta_proj, padding=padding)

    t1 = time.time()
    generator = bdg.conditional_generator(
        Nx,
        Ny,
        symmetry,
        periodic,
        padding,
        int(np.ceil(1.5 * brush_size)) | 1,
        maxiter=20,
    )

    x_reward = torch.tensor(x_reward, device='cuda' if torch.cuda.is_available() else 'cpu', dtype=torch.float64)
    x_near_binary = generator.generate_near_binary_no_grad(x_reward)
    x_bin_bdg = np.where(x_near_binary.detach().cpu().numpy() < 0.5, -1, 1).reshape(N_designs, Nx, Ny).astype(np.float32)
    t2 = time.time()
    if print_runtime_details:
        if comm.rank == 0:
            print("--> Bilevel Generator Runtime: " + str(t2 - t1) + " s", flush=True)

    t1 = time.time()
    try:
        if brush_size is not None:
            if Nthreads > 0:
                if upsample_ratio == 1:
                    x_brush = FDG.make_feasible_parallel(
                        x_bin_bdg,
                        brush_size,
                        periodic,
                        symmetry,
                        2,
                        upsample_ratio,
                        Nthreads).reshape(N_designs, -1)
                    
                else:
                    x_brush_lowres = FDG.make_feasible_parallel(
                        x_bin_bdg,
                        brush_size,
                        periodic,
                        symmetry,
                        2,
                        1,
                        Nthreads).reshape(N_designs, Nx, Ny)
                    
                    x_brush = FDG.make_feasible_parallel(
                        2*x_brush_lowres.astype(np.float32)-1,
                        brush_size,
                        periodic,
                        symmetry,
                        2,
                        upsample_ratio,
                        Nthreads).reshape(N_designs, -1)
            
            else:
                quo, rem = divmod(N_designs, comm.size)
                data_size = np.array([quo + 1 if p < rem else quo for p in range(comm.size)]).astype(np.int32)
                data_disp = np.array([sum(data_size[:p]) for p in range(comm.size + 1)])
                x_bin_bdg_proc = x_bin_bdg[data_disp[comm.rank]:data_disp[comm.rank+1]]
                
                if data_size[comm.rank] > 0:
                    if upsample_ratio == 1:
                        x_brush_proc = FDG.make_feasible_parallel(
                            x_bin_bdg_proc,
                            brush_size,
                            periodic,
                            symmetry,
                            2,
                            upsample_ratio,
                            1).reshape(data_size[comm.rank], -1).astype(np.float64)
                        
                    else:
                        x_brush_lowres_proc = FDG.make_feasible_parallel(
                            x_bin_bdg_proc,
                            brush_size,
                            periodic,
                            symmetry,
                            2,
                            1,
                            1).reshape(data_size[comm.rank], Nx, Ny)
                        
                        x_brush_proc = FDG.make_feasible_parallel(
                            2*x_brush_lowres_proc.astype(np.float32)-1,
                            brush_size,
                            periodic,
                            symmetry,
                            2,
                            upsample_ratio,
                            1).reshape(data_size[comm.rank], -1).astype(np.float64)
                else:
                    x_brush_proc = np.array([]).astype(np.float64)
                
                if padding is None:
                    data_size_temp = data_size*Nx*Ny*upsample_ratio**2
                    data_disp_temp = np.array([sum(data_size_temp[:p]) for p in range(comm.size)])
                    
                    x_brush_temp = np.zeros(N_designs*Nx*Ny*upsample_ratio**2)
                    comm.Allgatherv(x_brush_proc.reshape(-1), [x_brush_temp, data_size_temp, data_disp_temp, MPI.DOUBLE])
                    x_brush = x_brush_temp.reshape(N_designs, -1)
                else:
                    data_size_temp = data_size*(Nx + 2*brush_size)*(Ny + 2*brush_size)*upsample_ratio**2
                    data_disp_temp = np.array([sum(data_size_temp[:p]) for p in range(comm.size)])
                    
                    x_brush_temp = np.zeros(N_designs*(Nx + 2*brush_size)*(Ny + 2*brush_size)*upsample_ratio**2)
                    comm.Allgatherv(x_brush_proc.reshape(-1), [x_brush_temp, data_size_temp, data_disp_temp, MPI.DOUBLE])
                    x_brush = x_brush_temp.reshape(N_designs, -1)
    except:
        if comm.rank == 0:
            np.savez(os.path.join(os.path.expanduser("~"), "debug_brush"), x_bin_bdg=x_bin_bdg)
        assert False
    
    t2 = time.time()
    if print_runtime_details:
        if comm.rank == 0:
            print("--> Brush Generator Runtime: " + str(t2 - t1) + " s", flush=True)
    
    if padding is not None:
        x_brush_crop = x_brush.reshape(N_designs, (Nx + 2*brush_size)*upsample_ratio, (Ny + 2*brush_size)*upsample_ratio)
        x_brush_crop = x_brush_crop[:,brush_size:-brush_size,brush_size:-brush_size]
        x_brush = x_brush_crop.reshape(N_designs, -1)
    
    if output_details:
        return np.squeeze(x_sym), np.squeeze(x_brush)
    
    else:
        return np.squeeze(x_brush)


import numpy as np
import torch
from torch.func import jacrev
import gegd.parameter_processing.bilevel_design_generator as bdg
import time
import os

from mpi4py import MPI
comm = MPI.COMM_WORLD

class optimizer:
    def __init__(
		self,
		Nx,
		Ny,
		Ntrial,
		symmetry,
		periodic,
		padding,
		maxiter,
		high_fidelity_setting,
		brush_size,
		upsample_ratio=1,
		beta_proj=8,
		cost_obj=None,
		Nthreads=1,
	):
        
        self.Nx = Nx
        self.Ny = Ny
        self.Ntrial = Ntrial
        self.symmetry = symmetry
        self.periodic = periodic
        self.padding = padding
        self.maxiter = maxiter
        self.beta_proj = beta_proj
        self.beta_proj_sigma = beta_proj
        self.brush_size = brush_size
        self.upsample_ratio = upsample_ratio
        self.sigma_filter = brush_size/2/np.sqrt(2)
        self.high_fidelity_setting = high_fidelity_setting
        self.cost_obj = cost_obj
        self.Nthreads = Nthreads
        
        # Get Number of Independent Parameters
        if symmetry == 0:
            self.Ndim = Nx*Ny
           
        elif symmetry == 1:
            self.Ndim = int(np.floor(Nx/2 + 0.5)*Ny)
        
        elif symmetry == 2:
            self.Ndim = int(np.floor(Nx/2 + 0.5)*np.floor(Ny/2 + 0.5))
        
        elif symmetry == 4:
            self.Ndim = int(np.floor(Nx/2 + 0.5)*(np.floor(Nx/2 + 0.5) + 1)/2)
	
    def bilevel_cost(self, x_reward):
        # Get Bilevel Binarized Densities ------------------------------------------------------------
        generator = bdg.conditional_generator(
			self.Nx,
			self.Ny,
			self.symmetry,
			self.periodic,
			self.padding,
			self.brush_size + 2,
		)

        x_bin = np.zeros((self.Ntrial, self.Nx*self.Ny))
        f = np.zeros(self.Ntrial)
        jac = np.zeros((self.Ntrial, self.Ndim))

        for n in range(self.Ntrial):
            x_reward_temp = torch.tensor(x_reward[n,:], device='cuda' if torch.cuda.is_available() else 'cpu', dtype=torch.float64)
            jac_near_binary, x_near_binary = jacrev(generator.generate_near_binary, has_aux=True)(x_reward_temp)
            x_bin[n,:] = np.where(x_near_binary.detach().cpu().numpy() < 0.5, 0, 1).reshape(-1)

            # Sample Cost Function --------------------------------------------------------------
            t1 = time.time()
            f[n], jac_temp = self.cost_obj.get_cost(x_bin[n,:], get_grad=True)
            t2 = time.time()
            #print(t2 - t1, flush=True)

            jac[n,:] = jac_temp.T @ jac_near_binary.detach().cpu().numpy()
        
        return f, jac, x_bin

    def ADAM(self,
             x,
             beta_ADAM1,
             beta_ADAM2,
             eta_ADAM,
             jac_mean=None,
             jac_var=None,
             adam_iter=None,
             ):
        
        if jac_mean is None:
            jac_mean = np.zeros_like(x)
        if jac_var is None:
            jac_var = np.zeros_like(x)
        if adam_iter is None:
            adam_iter = 0
            
        while True:
            t1 = time.time()
            adam_iter += 1
            
            loss, jac, x_bin = self.bilevel_cost(x)

            best_sample = np.argmin(loss)

            if self.best_x_latent_hist is None:
                self.best_x_latent_hist = x[best_sample,:].copy()
            else:
                self.best_x_latent_hist = np.vstack((self.best_x_latent_hist, x[best_sample,:]))
                
            if self.best_x_hist is None:
                self.best_x_hist = x_bin[best_sample,:].copy()
            else:
                self.best_x_hist = np.vstack((self.best_x_hist, x_bin[best_sample,:]))
            
            if self.cost_hist is None:
                self.cost_hist = loss.copy()
            else:
                self.cost_hist = np.vstack((self.cost_hist, loss))
        
            t2 = time.time()
            t_rem = (t2 - t1)*(self.maxiter - self.n_iter + 1)/3600
            
            if comm.rank == 0:
                print('    | %4d |  %8.3f |   %5.2f   |' %(
                    self.n_iter,
                    np.min(loss),
                    t_rem,
                    ),
                    end='',
					flush=True,
				)

            self.save_data(
				x=x,
				jac_mean=jac_mean,
				jac_var=jac_var,
				adam_iter=adam_iter,
			)

            if adam_iter >= self.maxiter:
                break
                
            # Update Average Gradients
            jac_mean = beta_ADAM1*jac_mean + (1 - beta_ADAM1)*jac
            jac_var = beta_ADAM2*jac_var + (1 - beta_ADAM2)*jac**2
            
            # Unbias Average Gradients
            jac_mean_unbiased = jac_mean/(1 - beta_ADAM1**adam_iter)
            jac_var_unbiased = jac_var/(1 - beta_ADAM2**adam_iter)
            
            x -= eta_ADAM*jac_mean_unbiased/(np.sqrt(jac_var_unbiased) + 1e-8)
            
            self.n_iter += 1
		
    def run(self, n_seed, output_filename, x0=None, eta=0.01, load_data=False):
        if comm.rank == 0:
            print('### Bilevel Optimization (seed = ' + str(n_seed) + ')\n', flush=True)
    
        np.random.seed(n_seed)
        self.output_filename = output_filename
        
        if load_data:
            data_file1 = output_filename + "_AF_BL_results.npz"
            data_file2 = output_filename + "_AF_BL_density_hist.npz"
                
            with np.load(data_file1) as data:
                self.n_iter = data['n_iter']
                self.best_cost_hist = data['best_cost_hist'][:self.n_iter]

                adam_iter = data['adam_iter'] - 1
                
            with np.load(data_file2) as data:
                self.best_x_latent_hist = data['best_x_latent_hist'][:self.n_iter,:]
                self.best_x_hist = data['best_x_hist'][:self.n_iter,:]
                
                x0 = data['x_bounded']
                jac_mean = data['jac_mean']
                jac_var = data['jac_var']
            
        else:
            self.n_iter = 0
            
            self.best_x_latent_hist = None
            self.best_x_hist = None
            self.best_cost_hist = np.zeros(0)
            
            if x0 is None:
                # Initial Structure
                x0 = np.random.uniform(-1, 1, size=(self.Ntrial, self.Ndim))
            
            jac_mean = None
            jac_var = None
            adam_iter = None
        
        if comm.rank == 0:
            print('    | Iter | Cost Best | t_rem(hr) |', flush=True)

        eta_ADAM = eta*np.ones(self.Ndim)
        self.ADAM(
            x0,
			0.9,
			0.999,
			eta_ADAM,
            jac_mean=jac_mean,
			jac_var=jac_var,
			adam_iter=adam_iter,
		)
        
        self.save_data()
    
    def save_data(self, x_bounded=None, jac_mean=None, jac_var=None, adam_iter=None):
        if comm.rank == 0:
            if x_bounded is None:
                with np.load(self.output_filename + "_AF_BL_results.npz") as data:
                    adam_iter = data['adam_iter']
                
                with np.load(self.output_filename + "_AF_BL_density_hist.npz") as data:
                    x_bounded = data['x_bounded']
                    jac_mean = data['jac_mean']
                    jac_var = data['jac_var']
        
            # Customize below
            np.savez(self.output_filename + "_AF_BL_results",
                best_cost_hist=self.best_cost_hist,
                n_iter=self.n_iter,
                adam_iter=adam_iter,
            )
						
            np.savez(self.output_filename + "_AF_BL_density_hist",
                best_x_latent_hist=self.best_x_latent_hist,
                best_x_hist=self.best_x_hist,
                x_bounded=x_bounded,
                jac_mean=jac_mean,
                jac_var=jac_var,
			)



import sys
import os
import torch
import numpy as np

# Adjust path to import necessary modules
current_dir = os.path.dirname(os.path.abspath(__file__))
parent_dir = os.path.dirname(os.path.dirname(current_dir))
sys.path.append(parent_dir)

from gegd.parameter_processing.bilevel_design_generator import conditional_generator, filter_and_project

def verify_batch_consistency():
    Nx = 10
    Ny = 10
    symmetry = 0
    periodic = False
    mfs = 1
    padding = -np.ones((Nx + 2*mfs, Ny + 2*mfs))
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    
    print(f"Running on device: {device}")
    
    # Create two random inputs
    np.random.seed(42)
    x1_np = filter_and_project(np.random.uniform(-1, 1, size=(Nx*Ny)), periodic, Nx, Ny, mfs/2/np.sqrt(2), 8)/10
    x2_np = filter_and_project(np.random.uniform(-1, 1, size=(Nx*Ny)), periodic, Nx, Ny, mfs/2/np.sqrt(2), 8)/10
    
    x1 = torch.tensor(x1_np, device=device, dtype=torch.float64)
    x2 = torch.tensor(x2_np, device=device, dtype=torch.float64)
    
    # SETUP 1: Single processing
    print("Running single processing...")
    gen_single = conditional_generator(Nx, Ny, symmetry, periodic, padding, mfs + 2, x1, batch_processing=False)
    
    out1_single = gen_single.generate_near_binary_no_grad(x1)
    out2_single = gen_single.generate_near_binary_no_grad(x2)
    
    # SETUP 2: Batch processing
    print("Running batch processing...")
    batch_input = torch.stack([x1, x2])
    gen_batch = conditional_generator(Nx, Ny, symmetry, periodic, padding, mfs + 2, batch_input, batch_processing=True)
    
    out_batch = gen_batch.generate_near_binary_no_grad(batch_input)
    
    print(f"Batch output shape: {out_batch.shape}")
    
    # Compare results
    diff1 = torch.max(torch.abs(out_batch[0] - out1_single))
    diff2 = torch.max(torch.abs(out_batch[1] - out2_single))
    
    print(f"Max diff sample 1: {diff1}")
    print(f"Max diff sample 2: {diff2}")
    
    if diff1 < 1e-6 and diff2 < 1e-6:
        print("VERIFICATION PASSED: Batch processing matches single processing.")
    else:
        print("VERIFICATION FAILED: Results diverge.")

if __name__ == "__main__":
    verify_batch_consistency()



import numpy as np
import gegd.parameter_processing.density_transforms as dtf
import gegd.parameter_processing.symmetry_operations as symOp
import time

from mpi4py import MPI
comm = MPI.COMM_WORLD

class optimizer:
    def __init__(self,
                 Nx,
                 Ny,
                 Ntrial,
                 symmetry,
                 periodic,
                 padding,
                 high_fidelity_setting,
                 brush_size,
                 upsample_ratio=1,
                 beta_proj=8,
                 brush_shape='circle',
                 cost_obj=None,
                 Nthreads=1,
                 ):
                       
        self.Nx = Nx
        self.Ny = Ny
        self.Ntrial = Ntrial
        self.symmetry = symmetry
        self.periodic = periodic
        self.padding = padding
        self.beta_proj = beta_proj
        self.brush_size = brush_size
        self.brush_shape = brush_shape
        self.high_fidelity_setting = high_fidelity_setting
        self.cost_obj = cost_obj
        self.Nthreads = Nthreads
        
        # Get Number of Independent Parameters
        if symmetry == 0:
            self.Ndim = Nx*Ny
           
        elif symmetry == 1:
            self.Ndim = int(np.floor(Nx/2 + 0.5)*Ny)
        
        elif symmetry == 2:
            self.Ndim = int(np.floor(Nx/2 + 0.5)*np.floor(Ny/2 + 0.5))
        
        elif symmetry == 4:
            self.Ndim = int(np.floor(Nx/2 + 0.5)*(np.floor(Nx/2 + 0.5) + 1)/2)
    
    def straight_through_jacobian(self, x0):
        # Get Brush Binarized Densities ------------------------------------------------------------
        x_bin = dtf.binarize(x0, self.symmetry, self.periodic, self.Nx, self.Ny, self.brush_size, self.brush_shape, self.beta_proj, self.sigma_filter, padding=self.padding, Nthreads=self.Nthreads)

        # Sample Modified Cost Function --------------------------------------------------------------
        self.cost_obj.set_accuracy(self.high_fidelity_setting)

        f0 = np.zeros(self.Ntrial)
        jac_STE = np.zeros((self.Ntrial, self.Ndim))
        for n in range(self.Ntrial):
            f0[n], jac_temp = self.cost_obj.get_cost(x_bin[n,:], get_grad=True)
            jac_sym = jac_temp.reshape(self.Nx, self.Ny)
            jac_STE[n,:] = dtf.backprop_filter_and_project(jac_sym, x0[n,:], self.symmetry, self.periodic, self.Nx, self.Ny, self.brush_size, self.sigma_filter, self.beta_proj, padding=self.padding)
        
        return f0, jac_STE, x_bin
    
    def ADAM(self,
             x_bounded,
             lb,
             ub,
             beta_ADAM1,
             beta_ADAM2,
             eta_ADAM,
             maxiter,
             beta=0,
             jac_mean=None,
             jac_var=None,
             adam_iter=None,
             ):
    
        # Dummy Variables
        x = -np.log((ub[np.newaxis,:] - lb[np.newaxis,:])/(x_bounded - lb[np.newaxis,:]) - 1)
        x[:,ub==lb] = ub[ub==lb][np.newaxis,:]
        
        if jac_mean is None:
            jac_mean = np.zeros_like(x)
        if jac_var is None:
            jac_var = np.zeros_like(x)
        if adam_iter is None:
            adam_iter = 0
            
        while True:
            t1 = time.time()
            adam_iter += 1
            
            x_bounded = lb[np.newaxis,:] + (ub[np.newaxis,:] - lb[np.newaxis,:])/(1 + np.exp(-x))

            loss, jac, x_bin = self.straight_through_jacobian(x_bounded)

            jac[:,ub==lb] = 0
            jac *= np.exp(-x)*(ub[np.newaxis,:] - lb[np.newaxis,:])/(1 + np.exp(-x))**2

            if self.cost_hist is None:
                self.cost_hist = loss.copy()
            else:
                self.cost_hist = np.vstack((self.cost_hist, loss))
            
            if self.x_latent_hist is None:
                self.x_latent_hist = x_bounded[np.argmin(loss),:].copy()
            else:
                self.x_latent_hist = np.vstack((self.x_latent_hist, x_bounded[np.argmin(loss),:]))
            
            if self.x_hist is None:
                self.x_hist = x_bin[np.argmin(loss),:].copy()
            else:
                self.x_hist = np.vstack((self.x_hist, x_bin[np.argmin(loss),:]))

            t2 = time.time()

            if comm.rank == 0:
                t_rem = (t2 - t1)*(maxiter - self.n_iter + 1)/3600
                print('    | %12d | %12.5f |   %5.2f   |' %(self.n_iter, np.min(loss), t_rem), flush=True)

            self.save_data(x_bounded=x_bounded,
                           x_bin=x_bin,
                           jac_mean=jac_mean,
                           jac_var=jac_var,
                           adam_iter=adam_iter)

            if adam_iter >= maxiter:
                break

            # Update Average Gradients
            jac_mean = beta_ADAM1*jac_mean + (1 - beta_ADAM1)*jac
            jac_var = beta_ADAM2*jac_var + (1 - beta_ADAM2)*jac**2
            
            # Unbias Average Gradients
            jac_mean_unbiased = jac_mean/(1 - beta_ADAM1**adam_iter)
            jac_var_unbiased = jac_var/(1 - beta_ADAM2**adam_iter)
            
            # Update Variables
            x -= eta_ADAM*jac_mean_unbiased/(np.sqrt(jac_var_unbiased) + 1e-8)
            # eta *= 0.95

            self.n_iter += 1

        return x_bounded
    
    def run(self, n_seed, output_filename, maxiter, eta_ADAM=0.1, load_data=False):
        if comm.rank == 0:
            print('### Brush Optimization (seed = ' + str(n_seed) + ')\n', flush=True)
    
        self.output_filename = output_filename
        self.sigma_filter = self.brush_size/2/np.sqrt(2)

        lb = -np.ones(self.Ndim)
        ub = np.ones(self.Ndim)

        if load_data:
            data_file1 = output_filename + "_AF_STE_results.npz"
            data_file2 = output_filename + "_AF_STE_density_hist.npz"
                
            with np.load(data_file1) as data:
                self.n_iter = data['n_iter']
                self.cost_hist = data['cost_hist'][:self.n_iter,:]
                adam_iter = data['adam_iter'] - 1
                
            with np.load(data_file2) as data:
                self.x_latent_hist = data['x_latent_hist'][:self.n_iter,:]
                self.x_hist = data['x_hist'][:self.n_iter,:]
                x0 = data['x_bounded']
                jac_mean = data['jac_mean']
                jac_var = data['jac_var']
            
        else:
            self.x_latent_hist = None
            self.x_hist = None
            self.cost_hist = None
            
            self.n_iter = 0
        
            # Initial Structure
            if n_seed is not None:
                np.random.seed(n_seed)
            
            x0 = 2*np.random.rand(self.Ntrial, self.Ndim) - 1
            np.random.seed()
            
            jac_mean = None
            jac_var = None
            adam_iter = None

        if comm.rank == 0:
            print('    |  Iteration   |  Best Cost   | t_rem(hr) |', flush=True)
        
        x_bin = self.ADAM(x0, lb, ub, 0.667, 0.9, eta_ADAM, maxiter, jac_mean=jac_mean, jac_var=jac_var, adam_iter=adam_iter)
        
        if comm.rank == 0:
            print('', flush=True)
    
        self.save_data(x_bin=x_bin)
    
    def save_data(self, x_bounded=0, x_bin=0, jac_mean=0, jac_var=0, adam_iter=0):
        if comm.rank == 0:
            np.savez(self.output_filename + "_AF_STE_results",
                     cost_hist=self.cost_hist,
                     n_iter=self.n_iter,
                     adam_iter=adam_iter)
                     
            np.savez(self.output_filename + "_AF_STE_density_hist",
                     x_hist=self.x_hist,
                     x_latent_hist=self.x_latent_hist,
                     x_bounded=x_bounded,
                     x_bin=x_bin,
                     jac_mean=jac_mean,
                     jac_var=jac_var)


import numpy as np

def symmetrize(x, symmetry, Nx, Ny):
    if symmetry == 0:
        x_sym = x.reshape(Nx, Ny)
    
    elif symmetry == 1:
        x_half = x.reshape(int(np.floor(Nx/2 + 0.5)), Ny)
        x_sym = np.zeros((Nx, Ny))
        
        if Nx % 2 == 0:
            x_sym[:int(Nx/2),:] = x_half
            x_sym[int(Nx/2):,:] = np.flip(x_half, axis=0)
        
        else:
            x_sym[:int((Nx+1)/2),:] = x_half
            x_sym[int((Nx+1)/2):,:] = np.flip(x_half[:-1,:], axis=0)
    
    elif symmetry == 2:
        x_quart = x.reshape(int(np.floor(Nx/2 + 0.5)), int(np.floor(Ny/2 + 0.5)))
        x_sym = np.zeros((Nx, Ny))
        
        if Nx % 2 == 0 and Ny % 2 == 0:
            x_sym[:int(Nx/2),:int(Ny/2)] = x_quart
            x_sym[int(Nx/2):,:int(Ny/2)] = np.flip(x_quart, axis=0)
            x_sym[:int(Nx/2),int(Ny/2):] = np.flip(x_quart, axis=1)
            x_sym[int(Nx/2):,int(Ny/2):] = np.flip(np.flip(x_quart, axis=0), axis=1)
        
        elif Nx % 2 == 1 and Ny % 2 == 0:
            x_sym[:int((Nx+1)/2),:int(Ny/2)] = x_quart
            x_sym[int((Nx+1)/2):,:int(Ny/2)] = np.flip(x_quart[:-1,:], axis=0)
            x_sym[:int((Nx+1)/2),int(Ny/2):] = np.flip(x_quart, axis=1)
            x_sym[int((Nx+1)/2):,int(Ny/2):] = np.flip(np.flip(x_quart[:-1,:], axis=0), axis=1)
        
        elif Nx % 2 == 0 and Ny % 2 == 1:
            x_sym[:int(Nx/2),:int((Ny+1)/2)] = x_quart
            x_sym[int(Nx/2):,:int((Ny+1)/2)] = np.flip(x_quart, axis=0)
            x_sym[:int(Nx/2),int((Ny+1)/2):] = np.flip(x_quart[:,:-1], axis=1)
            x_sym[int(Nx/2):,int((Ny+1)/2):] = np.flip(np.flip(x_quart[:,:-1], axis=0), axis=1)
        
        elif Nx % 2 == 1 and Ny % 2 == 1:
            x_sym[:int((Nx+1)/2),:int((Ny+1)/2)] = x_quart
            x_sym[int((Nx+1)/2):,:int((Ny+1)/2)] = np.flip(x_quart[:-1,:], axis=0)
            x_sym[:int((Nx+1)/2),int((Ny+1)/2):] = np.flip(x_quart[:,:-1], axis=1)
            x_sym[int((Nx+1)/2):,int((Ny+1)/2):] = np.flip(np.flip(x_quart[:-1,:-1], axis=0), axis=1)
    
    elif symmetry == 4:
        assert Nx == Ny, "The pixel numbers in the X and Y directions must be equal under D4 symmetry"
        
        x = x.reshape(-1)            
        if Nx % 2 == 0:
            x_quart = np.zeros((int(Nx/2), int(Ny/2)))
            x_quart[np.triu_indices(int(Nx/2))] = x
            x_quart = x_quart + x_quart.T - np.diag(np.diag(x_quart))
            
            x_sym = np.zeros((Nx, Ny))
            x_sym[:int(Nx/2),:int(Ny/2)] = x_quart
            x_sym[int(Nx/2):,:int(Ny/2)] = np.rot90(x_quart)
            x_sym[int(Nx/2):,int(Ny/2):] = np.rot90(x_quart, k=2)
            x_sym[:int(Nx/2),int(Ny/2):] = np.rot90(x_quart, k=-1)
        
        else:
            x_quart = np.zeros((int((Nx+1)/2), int((Ny+1)/2)))
            x_quart[np.triu_indices(int((Nx+1)/2))] = x
            x_quart = x_quart + x_quart.T - np.diag(np.diag(x_quart))
            
            x_sym = np.zeros((Nx, Ny))
            x_sym[:int((Nx+1)/2),:int((Ny+1)/2)] = x_quart
            x_sym[int((Nx+1)/2):,:int((Ny+1)/2)] = np.rot90(x_quart)[1:,:]
            x_sym[int((Nx+1)/2):,int((Ny+1)/2):] = np.rot90(x_quart, k=2)[1:,1:]
            x_sym[:int((Nx+1)/2),int((Ny+1)/2):] = np.rot90(x_quart, k=-1)[:,1:]
    
    return x_sym

def symmetrize_jacobian(jac, symmetry, Nx, Ny):
    if symmetry == 0:
        jac_sym = symmetrize(jac, symmetry, Nx, Ny)
    
    elif symmetry == 1:
        if Nx % 2 == 0:
            scaling_coeff = (1/2)*np.ones((int(Nx/2), Ny))
            
        else:
            scaling_coeff = (1/2)*np.ones((int((Nx+1)/2), Ny))
            scaling_coeff[-1,:] = 1
        
        jac *= scaling_coeff.reshape(-1)
        jac_sym = symmetrize(jac, symmetry, Nx, Ny)
    
    elif symmetry == 2:
        if Nx % 2 == 0 and Ny % 2 == 0:
            scaling_coeff = (1/4)*np.ones((int(Nx/2), int(Ny/2)))
        
        elif Nx % 2 == 1 and Ny % 2 == 0:
            scaling_coeff = (1/4)*np.ones((int((Nx+1)/2), int(Ny/2)))
            scaling_coeff[-1,:] = 1/2
        
        elif Nx % 2 == 0 and Ny % 2 == 1:
            scaling_coeff = (1/4)*np.ones((int(Nx/2), int((Ny+1)/2)))
            scaling_coeff[:,-1] = 1/2
        
        else:
            scaling_coeff = (1/4)*np.ones((int((Nx+1)/2), int((Ny+1)/2)))
            scaling_coeff[-1,:] = 1/2
            scaling_coeff[:,-1] = 1/2
            scaling_coeff[-1,-1] = 1
        
        jac *= scaling_coeff.reshape(-1)
        jac_sym = symmetrize(jac, symmetry, Nx, Ny)
    
    elif symmetry == 4:
        if Nx % 2 == 0:
            scaling_coeff = (1/8)*np.ones((int(Nx/2), int(Ny/2)))
            scaling_coeff += (1/8)*np.identity(int(Nx/2))
            scaling_coeff = scaling_coeff[np.triu_indices(int(Nx/2))]
        
        else:
            scaling_coeff = (1/8)*np.ones((int((Nx+1)/2), int((Ny+1)/2)))
            scaling_coeff += (1/8)*np.identity(int((Nx+1)/2))
            scaling_coeff[:,-1] = 1/4
            scaling_coeff[-1,-1] = 1
            scaling_coeff = scaling_coeff[np.triu_indices(int((Nx+1)/2))]
        
        jac *= scaling_coeff
        jac_sym = symmetrize(jac, symmetry, Nx, Ny)
    
    return jac_sym

def desymmetrize(x_sym, symmetry, Nx, Ny):
    if symmetry == 0:
        x = x_sym.reshape(-1)
    
    elif symmetry == 1:
        x_sym = x_sym.reshape(Nx, Ny)
        if Nx % 2 == 0:
            x_half = x_sym[:int(Nx/2),:]
        else:
            x_half = x_sym[:int((Nx+1)/2),:]
            
        x = x_half.reshape(-1)
    
    elif symmetry == 2:
        x_sym = x_sym.reshape(Nx, Ny)
        if Nx % 2 == 0 and Ny % 2 == 0:
            x_quart = x_sym[:int(Nx/2),:int(Ny/2)]
        elif Nx % 2 == 1 and Ny % 2 == 0:
            x_quart = x_sym[:int((Nx+1)/2),:int(Ny/2)]
        elif Nx % 2 == 0 and Ny % 2 == 1:
            x_quart = x_sym[:int(Nx/2),:int((Ny+1)/2)]
        else:
            x_quart = x_sym[:int((Nx+1)/2),:int((Ny+1)/2)]
        
        x = x_quart.reshape(-1)
    
    elif symmetry == 4:
        x_sym = x_sym.reshape(Nx, Ny)
        if Nx % 2 == 0:
            x_quart = x_sym[:int(Nx/2),:int(Ny/2)]
            x = x_quart[np.triu_indices(int(Nx/2))]
        
        else:
            x_quart = x_sym[:int((Nx+1)/2),:int((Ny+1)/2)]
            x = x_quart[np.triu_indices(int((Nx+1)/2))]
    
    return x
    
def desymmetrize_jacobian(jac_sym, symmetry, Nx, Ny):
    if symmetry == 0:
        jac = jac_sym.copy()
    
    elif symmetry == 1:
        if Nx % 2 == 0:
            jac = jac_sym[:int(Nx/2),:]
            jac += np.flip(jac_sym[int(Nx/2):,:], axis=0)
        
        else:
            jac = jac_sym[:int((Nx+1)/2),:]
            jac[:-1,:] += np.flip(jac_sym[int((Nx+1)/2):,:], axis=0)
    
    elif symmetry == 2:
        if Nx % 2 == 0 and Ny % 2 == 0:
            jac = jac_sym[:int(Nx/2),:int(Ny/2)]
            jac += np.flip(jac_sym[int(Nx/2):,:int(Ny/2)], axis=0)
            jac += np.flip(jac_sym[:int(Nx/2),int(Ny/2):], axis=1)
            jac += np.flip(jac_sym[int(Nx/2):,int(Ny/2):], axis=(0,1))
        
        elif Nx % 2 == 1 and Ny % 2 == 0:
            jac = jac_sym[:int((Nx+1)/2),:int(Ny/2)]
            jac[:-1,:] += np.flip(jac_sym[int((Nx+1)/2):,:int(Ny/2)], axis=0)
            jac += np.flip(jac_sym[:int((Nx+1)/2),int(Ny/2):], axis=1)
            jac[:-1,:] += np.flip(jac_sym[int((Nx+1)/2):,int(Ny/2):], axis=(0,1))
        
        elif Nx % 2 == 0 and Ny % 2 == 1:
            jac = jac_sym[:int(Nx/2),:int((Ny+1)/2)]
            jac += np.flip(jac_sym[int(Nx/2):,:int((Ny+1)/2)], axis=0)
            jac[:,:-1] += np.flip(jac_sym[:int(Nx/2),int((Ny+1)/2):], axis=1)
            jac[:,:-1] += np.flip(jac_sym[int(Nx/2):,int((Ny+1)/2):], axis=(0,1))
        
        else:
            jac = jac_sym[:int((Nx+1)/2),:int((Ny+1)/2)]
            jac[:-1,:] = np.flip(jac_sym[int((Nx+1)/2):,:int((Ny+1)/2)], axis=0)
            jac[:,:-1] = np.flip(jac_sym[:int((Nx+1)/2),int((Ny+1)/2):], axis=1)
            jac[:-1,:-1] = np.flip(jac_sym[int((Nx+1)/2):,int((Ny+1)/2):], axis=(0,1))
    
    elif symmetry == 4:
        if Nx % 2 == 0:
            jac_quart = jac_sym[:int(Nx/2),:int(Ny/2)]
            jac_quart += np.flip(jac_sym[int(Nx/2):,:int(Ny/2)], axis=0)
            jac_quart += np.flip(jac_sym[:int(Nx/2),int(Ny/2):], axis=1)
            jac_quart += np.flip(jac_sym[int(Nx/2):,int(Ny/2):], axis=(0,1))
            jac_quart += jac_quart.T - np.diag(np.diag(jac_quart))
            
            jac = jac_quart[np.triu_indices(int(Nx/2))]

        else:
            jac_quart = jac_sym[:int((Nx+1)/2),:int((Ny+1)/2)]
            jac_quart[:-1,:] = np.flip(jac_sym[int((Nx+1)/2):,:int((Ny+1)/2)], axis=0)
            jac_quart[:,:-1] = np.flip(jac_sym[:int((Nx+1)/2),int((Ny+1)/2):], axis=1)
            jac_quart[:-1,:-1] = np.flip(jac_sym[int((Nx+1)/2):,int((Ny+1)/2):], axis=(0,1))
            jac_quart += jac_quart.T - np.diag(np.diag(jac_quart))
            
            jac = jac_quart[np.triu_indices(int((Nx+1)/2))]
    
    return jac.reshape(-1)